@@ -2,7 +2,7 @@
 # This file is part of the bumpver project
 # https://github.com/mbarkhau/bumpver
 #
-# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
+# Copyright (c) 2018-2024 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
 # SPDX-License-Identifier: MIT
 """cli module for BumpVer."""
 import io
@@ -37,7 +37,7 @@ except ImportError:
     pass  # no need to fail because of missing dev dependency
 
 
-click.disable_unicode_literals_warning = True  # type: ignore[attr-defined]
+click.disable_unicode_literals_warning = True
 
 logger = logging.getLogger("bumpver.cli")
 
@@ -68,7 +68,7 @@ def _configure_logging(verbose: int = 0) -> None:
     logger.debug("Logging configured.")
 
 
-VALID_RELEASE_TAG_VALUES = ("alpha", "beta", "rc", "post", "final")
+VALID_RELEASE_TAG_VALUES = ("alpha", "beta", "dev", "rc", "post", "final")
 
 
 _current_date = dt.date.today().isoformat()
@@ -86,9 +86,7 @@ def _validate_date(date: typ.Optional[str], pin_date: bool) -> typ.Optional[dt.d
         dt_val = dt.datetime.strptime(date, "%Y-%m-%d")
         return dt_val.date()
     except ValueError:
-        logger.error(
-            f"Invalid parameter --date='{date}', must match format YYYY-0M-0D.", exc_info=True
-        )
+        logger.error(f"Invalid parameter --date='{date}', must match format YYYY-0M-0D.", exc_info=True)
         sys.exit(1)
 
 
@@ -181,6 +179,26 @@ dry_option = click.option(
 )
 
 
+allow_dirty_option = click.option(
+    "--allow-dirty",
+    default=False,
+    is_flag=True,
+    help=(
+        "Commit even when working directory is has uncomitted changes. "
+        "(WARNING: The commit will still be aborted if there are uncomitted "
+        "to files with version strings."
+    ),
+)
+
+
+ignore_vcs_tag_option = click.option(
+    "--ignore-vcs-tag",
+    default=False,
+    is_flag=True,
+    help="Ignore VCS tag invariant and update version anyway.",
+)
+
+
 fetch_option = click.option(
     "-f/-n",
     "--fetch/--no-fetch",
@@ -190,24 +208,31 @@ fetch_option = click.option(
 )
 
 
+# env_option is depricated in favour of --environ
+# see https://github.com/mbarkhau/bumpver/issues/224
 env_option = click.option(
     "-e",
     "--env",
     is_flag=True,
     default=False,
+    hidden=True,
     help="Print version state for use with shell scripts: eval $(bumpver show --env)",
 )
 
 
+environ_option = click.option(
+    "--environ",
+    is_flag=True,
+    default=False,
+    help="Print version state for use with shell scripts: eval $(bumpver show --environ)",
+)
+
+
 def version_options(function: typ.Callable) -> typ.Callable:
     decorators = [
         click.option("--major", is_flag=True, default=False, help="Increment MAJOR component."),
-        click.option(
-            "-m", "--minor", is_flag=True, default=False, help="Increment MINOR component."
-        ),
-        click.option(
-            "-p", "--patch", is_flag=True, default=False, help="Increment PATCH component."
-        ),
+        click.option("-m", "--minor", is_flag=True, default=False, help="Increment MINOR component."),
+        click.option("-p", "--patch", is_flag=True, default=False, help="Increment PATCH component."),
         click.option(
             "-t",
             "--tag",
@@ -225,8 +250,12 @@ def version_options(function: typ.Callable) -> typ.Callable:
             help="Increment release tag number (rc1, rc2, rc3..).",
         ),
         click.option(
-            "--pin-date", is_flag=True, default=False, help="Leave date components unchanged."
+            "--pin-increments",
+            is_flag=True,
+            default=False,
+            help="Leave the auto-increments INC0 and INC1 unchanged.",
         ),
+        click.option("--pin-date", is_flag=True, default=False, help="Leave date components unchanged."),
         click.option(
             "--date",
             default=None,
@@ -247,7 +276,7 @@ def version_options(function: typ.Callable) -> typ.Callable:
 
 
 @click.group(context_settings={'help_option_names': ["-h", "--help"]})
-@click.version_option(version="2022.1119")
+@click.version_option(version="2024.1130")
 @verbose_option
 def cli(verbose: int = 0) -> None:
     """Automatically update version strings in plaintext files."""
@@ -261,17 +290,18 @@ def cli(verbose: int = 0) -> None:
 @verbose_option
 @version_options
 def test(
-    old_version: str,
-    pattern    : str,
-    verbose    : int = 0,
-    major      : bool = False,
-    minor      : bool = False,
-    patch      : bool = False,
-    tag        : str = None,
-    tag_num    : bool = False,
-    pin_date   : bool = False,
-    date       : typ.Optional[str] = None,
-    set_version: typ.Optional[str] = None,
+    old_version   : str,
+    pattern       : str,
+    verbose       : int = 0,
+    major         : bool = False,
+    minor         : bool = False,
+    patch         : bool = False,
+    tag           : typ.Optional[str] = None,
+    tag_num       : bool = False,
+    pin_increments: bool = False,
+    pin_date      : bool = False,
+    date          : typ.Optional[str] = None,
+    set_version   : typ.Optional[str] = None,
 ) -> None:
     """Increment a version number for demo purposes."""
     _configure_logging(verbose=max(_VERBOSE, verbose))
@@ -291,6 +321,7 @@ def test(
             patch=patch,
             tag=tag,
             tag_num=tag_num,
+            pin_increments=pin_increments,
             pin_date=pin_date,
             maybe_date=maybe_date,
         )
@@ -332,9 +363,7 @@ def _grep_text(pattern: patterns.Pattern, text: str, color: bool) -> typ.Iterabl
             )
         else:
             matched_line = (
-                text[line_start:match_start]
-                + text[match_start:match_end]
-                + text[match_end:line_end]
+                text[line_start:match_start] + text[match_start:match_end] + text[match_end:line_end]
             )
 
         lines_offset = max(0, line_idx - 1) + 1
@@ -421,9 +450,17 @@ def grep(
 
 @cli.command()
 @verbose_option
+@ignore_vcs_tag_option
 @fetch_option
 @env_option
-def show(verbose: int = 0, fetch: bool = True, env: bool = False) -> None:
+@environ_option
+def show(
+    verbose       : int = 0,
+    ignore_vcs_tag: bool = False,
+    fetch         : bool = True,
+    env           : bool = False,
+    environ       : bool = False,
+) -> None:
     """Show current version of your project."""
     _configure_logging(verbose=max(_VERBOSE, verbose))
 
@@ -433,13 +470,23 @@ def show(verbose: int = 0, fetch: bool = True, env: bool = False) -> None:
         logger.error("Could not parse configuration. Perhaps try 'bumpver init'.")
         sys.exit(1)
 
-    cfg = _update_cfg_from_vcs(cfg, fetch)
+    if not ignore_vcs_tag:
+        cfg = _update_cfg_from_vcs(cfg, fetch)
+
     if env:
+        logger.warning("Depricated: -e/--env use --environ instead. ")
+        logger.warning("    See https://github.com/mbarkhau/bumpver/issues/224")
         version_info = v2version.parse_version_info(cfg.current_version, cfg.version_pattern)
         for key, val in version_info._asdict().items():
             click.echo(f"{key.upper()}={val if val else ''}")
         click.echo(f"CURRENT_VERSION={cfg.current_version}")
         click.echo(f"PEP440_VERSION={cfg.pep440_version}")
+    elif environ:
+        version_info = v2version.parse_version_info(cfg.current_version, cfg.version_pattern)
+        for key, val in version_info._asdict().items():
+            click.echo(f"{key.upper()}={'' if (val is False or val is None) else val}")
+        click.echo(f"CURRENT_VERSION={cfg.current_version}")
+        click.echo(f"PEP440_VERSION={cfg.pep440_version}")
     else:
         click.echo(f"Current Version: {cfg.current_version}")
         click.echo(f"PEP440         : {cfg.pep440_version}")
@@ -498,8 +545,14 @@ def _print_diff(cfg: config.Config, new_version: str) -> None:
         sys.exit(1)
 
 
-def _is_valid_version(raw_pattern: str, old_version: str, new_version: str) -> bool:
+def _parse_version_tags(all_tags: typ.List[str], version_pattern: str, is_new_pattern: bool) -> typ.List[str]:
+    version_parser = v2version if is_new_pattern else v1version
+    return [tag for tag in all_tags if version_parser.is_valid(tag, version_pattern)]
+
+
+def _is_valid_version(raw_pattern: str, old_version: str, new_version: str, unique: bool = False) -> bool:
     is_new_pattern = "{" not in raw_pattern and "}" not in raw_pattern
+
     try:
         if is_new_pattern:
             v2version.parse_version_info(new_version, raw_pattern)
@@ -512,22 +565,32 @@ def _is_valid_version(raw_pattern: str, old_version: str, new_version: str) -> b
     if version.parse_version(new_version) <= version.parse_version(old_version):
         logger.error("Invariant violated: New version must be greater than old version ")
         logger.error(f"  Failed Invariant: '{new_version}' > '{old_version}'")
+        logger.error("If the invariant is from vcs tags try '--ignore-vcs-tag' option.")
         return False
-    else:
-        return True
+
+    if unique:
+        all_tags     = vcs.get_tags(fetch=False, scope=config.TagScope.GLOBAL)
+        version_tags = _parse_version_tags(all_tags, raw_pattern, is_new_pattern)
+
+        if new_version in version_tags:
+            logger.error("Invariant violated: New version must be unique accross all branches")
+            return False
+
+    return True
 
 
 def incr_dispatch(
     old_version: str,
     raw_pattern: str,
     *,
-    major     : bool = False,
-    minor     : bool = False,
-    patch     : bool = False,
-    tag       : str = None,
-    tag_num   : bool = False,
-    pin_date  : bool = False,
-    maybe_date: typ.Optional[dt.date] = None,
+    major         : bool = False,
+    minor         : bool = False,
+    patch         : bool = False,
+    tag           : typ.Optional[str] = None,
+    tag_num       : bool = False,
+    pin_increments: bool = False,
+    pin_date      : bool = False,
+    maybe_date    : typ.Optional[dt.date] = None,
 ) -> typ.Optional[str]:
     v1_parts    = list(v1patterns.PART_PATTERNS) + list(v1patterns.FULL_PART_FORMATS)
     has_v1_part = any("{" + part + "}" in raw_pattern for part in v1_parts)
@@ -562,6 +625,7 @@ def incr_dispatch(
             patch=patch,
             tag=tag,
             tag_num=tag_num,
+            pin_increments=pin_increments,
             pin_date=pin_date,
             maybe_date=maybe_date,
         )
@@ -571,6 +635,7 @@ def _update(
     cfg           : config.Config,
     new_version   : str,
     commit_message: str,
+    tag_message   : str,
     allow_dirty   : bool = False,
 ) -> None:
     vcs_api: typ.Optional[vcs.VCSAPI] = None
@@ -598,17 +663,18 @@ def _update(
         sys.exit(1)
 
     if vcs_api:
-        vcs.commit(cfg, vcs_api, filepaths, new_version, commit_message)
+        vcs.commit(cfg, vcs_api, filepaths, new_version, commit_message, tag_message)
 
 
 def _try_update(
     cfg           : config.Config,
     new_version   : str,
     commit_message: str,
+    tag_message   : str,
     allow_dirty   : bool = False,
 ) -> None:
     try:
-        _update(cfg, new_version, commit_message, allow_dirty)
+        _update(cfg, new_version, commit_message, tag_message, allow_dirty)
     except sp.CalledProcessError as ex:
         logger.error(f"Error running subcommand: {ex.cmd}")
         if ex.stdout:
@@ -641,12 +707,8 @@ def init(verbose: int = 0, dry: bool = False) -> None:
 
 
 def get_latest_vcs_version_tag(cfg: config.Config, fetch: bool) -> typ.Optional[str]:
-    all_tags = vcs.get_tags(fetch=fetch)
-
-    if cfg.is_new_pattern:
-        version_tags = [tag for tag in all_tags if v2version.is_valid(tag, cfg.version_pattern)]
-    else:
-        version_tags = [tag for tag in all_tags if v1version.is_valid(tag, cfg.version_pattern)]
+    all_tags     = vcs.get_tags(fetch=fetch, scope=cfg.tag_scope)
+    version_tags = _parse_version_tags(all_tags, cfg.version_pattern, cfg.is_new_pattern)
 
     if version_tags:
         version_tags.sort(key=version.parse_version, reverse=True)
@@ -659,28 +721,36 @@ def get_latest_vcs_version_tag(cfg: config.Config, fetch: bool) -> typ.Optional[
 
 def _update_cfg_from_vcs(cfg: config.Config, fetch: bool) -> config.Config:
     latest_version_tag = get_latest_vcs_version_tag(cfg, fetch)
+
     if latest_version_tag is None:
         logger.debug("no vcs tags found")
         return cfg
-    else:
-        latest_version_pep440 = version.to_pep440(latest_version_tag)
+
+    latest_version_pep440 = version.to_pep440(latest_version_tag)
+
+    scope_str = f"({cfg.tag_scope.value})" if not cfg.tag_scope == config.TagScope.DEFAULT else ""
+    logger.info(f"Latest version from VCS tag: {latest_version_tag} {scope_str}")
+
+    if cfg.tag_scope == config.TagScope.DEFAULT:
+        logger.info(f"Working dir version        : {cfg.current_version}")
         if version.parse_version(latest_version_tag) <= version.parse_version(cfg.current_version):
             # current_version already newer/up-to-date
             return cfg
-        else:
-            logger.info(f"Working dir version        : {cfg.current_version}")
-            logger.info(f"Latest version from VCS tag: {latest_version_tag}")
-            return cfg._replace(
-                current_version=latest_version_tag,
-                pep440_version=latest_version_pep440,
-            )
+
+    return cfg._replace(
+        current_version=latest_version_tag,
+        pep440_version=latest_version_pep440,
+    )
 
 
 def _parse_vcs_options(
-    cfg       : config.Config,
-    commit    : typ.Optional[bool] = None,
-    tag_commit: typ.Optional[bool] = None,
-    push      : typ.Optional[bool] = None,
+    cfg             : config.Config,
+    commit          : typ.Optional[bool] = None,
+    tag_commit      : typ.Optional[bool] = None,
+    push            : typ.Optional[bool] = None,
+    tag_scope       : typ.Optional[str] = None,
+    pre_commit_hook : typ.Optional[str] = None,
+    post_commit_hook: typ.Optional[str] = None,
 ) -> config.Config:
     if commit is False and tag_commit:
         raise ValueError("--no-commit and --tag-commit cannot be used at the same time")
@@ -699,23 +769,26 @@ def _parse_vcs_options(
         cfg = cfg._replace(tag=tag_commit)
     if push is not None:
         cfg = cfg._replace(push=push)
+    if tag_scope is not None:
+        cfg = cfg._replace(tag_scope=config.TagScope(tag_scope))
+    if pre_commit_hook is not None:
+        cfg = cfg._replace(pre_commit_hook=pre_commit_hook)
+    if post_commit_hook is not None:
+        cfg = cfg._replace(post_commit_hook=post_commit_hook)
+
     return cfg
 
 
+def _sub_msg_template(message: str) -> str:
+    return re.sub(r"\b(OLD|NEW)\b", r"{\1_VERSION}", message)
+
+
 @cli.command()
 @dry_option
+@allow_dirty_option
+@ignore_vcs_tag_option
 @fetch_option
 @verbose_option
-@click.option(
-    "--allow-dirty",
-    default=False,
-    is_flag=True,
-    help=(
-        "Commit even when working directory is has uncomitted changes. "
-        "(WARNING: The commit will still be aborted if there are uncomitted "
-        "to files with version strings."
-    ),
-)
 @version_options
 @click.option(
     "-c",
@@ -724,6 +797,12 @@ def _parse_vcs_options(
     metavar="<TMPL>",
     help="Set commit message template.",
 )
+@click.option(
+    "--tag-message",
+    default=None,
+    metavar="<TMPL>",
+    help="Set tag message template.",
+)
 @click.option(
     "--commit/--no-commit",
     default=None,
@@ -739,23 +818,48 @@ def _parse_vcs_options(
     default=None,
     help="Push to the default remote.",
 )
+@click.option(
+    "--tag-scope",
+    default=None,
+    metavar="[default|global|branch]",
+    type=click.Choice([e.value for e in config.TagScope]),
+    help="Tag scope for the current version.",
+)
+@click.option(
+    "--pre-commit-hook",
+    type=click.Path(exists=True),
+    metavar="<PATH>",
+    help="Custom script that runs before the commit step",
+)
+@click.option(
+    "--post-commit-hook",
+    type=click.Path(exists=True),
+    metavar="<PATH>",
+    help="Custom script that runs after the commit step is completed",
+)
 def update(
-    dry           : bool = False,
-    allow_dirty   : bool = False,
-    fetch         : bool = True,
-    verbose       : int = 0,
-    major         : bool = False,
-    minor         : bool = False,
-    patch         : bool = False,
-    tag           : typ.Optional[str] = None,
-    tag_num       : bool = False,
-    pin_date      : bool = False,
-    date          : typ.Optional[str] = None,
-    set_version   : typ.Optional[str] = None,
-    commit_message: typ.Optional[str] = None,
-    commit        : typ.Optional[bool] = None,
-    tag_commit    : typ.Optional[bool] = None,
-    push          : typ.Optional[bool] = None,
+    dry             : bool = False,
+    allow_dirty     : bool = False,
+    ignore_vcs_tag  : bool = False,
+    fetch           : bool = True,
+    verbose         : int = 0,
+    major           : bool = False,
+    minor           : bool = False,
+    patch           : bool = False,
+    tag             : typ.Optional[str] = None,
+    tag_num         : bool = False,
+    pin_increments  : bool = False,
+    pin_date        : bool = False,
+    date            : typ.Optional[str] = None,
+    set_version     : typ.Optional[str] = None,
+    commit_message  : typ.Optional[str] = None,
+    tag_message     : typ.Optional[str] = None,
+    commit          : typ.Optional[bool] = None,
+    tag_commit      : typ.Optional[bool] = None,
+    push            : typ.Optional[bool] = None,
+    tag_scope       : typ.Optional[str] = None,
+    pre_commit_hook : typ.Optional[str] = None,
+    post_commit_hook: typ.Optional[str] = None,
 ) -> None:
     """Update project files with the incremented version string."""
     verbose = max(_VERBOSE, verbose)
@@ -770,12 +874,13 @@ def update(
         sys.exit(1)
 
     try:
-        cfg = _parse_vcs_options(cfg, commit, tag_commit, push)
+        cfg = _parse_vcs_options(cfg, commit, tag_commit, push, tag_scope, pre_commit_hook, post_commit_hook)
     except ValueError as ex:
         logger.warning(f"Invalid argument: {ex}")
         sys.exit(1)
 
-    cfg = _update_cfg_from_vcs(cfg, fetch)
+    if not ignore_vcs_tag:
+        cfg = _update_cfg_from_vcs(cfg, fetch)
 
     old_version = cfg.current_version
     if set_version is None:
@@ -787,6 +892,7 @@ def update(
             patch=patch,
             tag=tag,
             tag_num=tag_num,
+            pin_increments=pin_increments,
             pin_date=pin_date,
             maybe_date=maybe_date,
         )
@@ -797,10 +903,11 @@ def update(
         _log_no_change('update', cfg.version_pattern)
         sys.exit(1)
 
-    if not _is_valid_version(cfg.version_pattern, old_version, new_version):
+    uniqueness_check = cfg.tag_scope == config.TagScope.BRANCH or set_version is not None
+
+    if not _is_valid_version(cfg.version_pattern, old_version, new_version, unique=uniqueness_check):
         if set_version:
             logger.error(f"Invalid argument --set-version='{set_version}'")
-
         sys.exit(1)
 
     logger.info(f"Old Version: {old_version}")
@@ -812,9 +919,11 @@ def update(
     if commit_message is None:
         commit_msg_template = cfg.commit_message
     else:
-        commit_msg_template, _ = re.subn(r"\b(OLD|NEW)\b", r"{\1_VERSION}", commit_message)
+        commit_msg_template = _sub_msg_template(commit_message)
 
-    commit_message_kwargs = {
+    tag_msg_template = cfg.tag_message if tag_message is None else _sub_msg_template(tag_message)
+
+    tag_and_commit_message_kwargs = {
         'new_version'       : new_version,
         'old_version'       : old_version,
         'NEW_VERSION'       : new_version,
@@ -822,12 +931,14 @@ def update(
         'new_version_pep440': version.to_pep440(new_version),
         'old_version_pep440': version.to_pep440(old_version),
     }
-    try_commit_message = commit_msg_template.format(**commit_message_kwargs)
+
+    try_commit_message = commit_msg_template.format(**tag_and_commit_message_kwargs)
+    try_tag_message    = tag_msg_template.format(**tag_and_commit_message_kwargs)
 
     if dry:
         return
 
-    _try_update(cfg, new_version, try_commit_message, allow_dirty)
+    _try_update(cfg, new_version, try_commit_message, try_tag_message, allow_dirty)
 
 
 if __name__ == '__main__':
