@@ -1,19 +1,21 @@
 # This file is part of the bumpver project
 # https://gitlab.com/mbarkhau/bumpver
 #
-# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
+# Copyright (c) 2018-2024 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
 # SPDX-License-Identifier: MIT
 """Parse bumpver.toml, setup.cfg or pyproject.toml files."""
 
 import re
+import enum
 import typing as typ
 import logging
-import datetime as dt
 import configparser
 
 import toml
-import pathlib2 as pl
 
+from bumpver import pathlib as pl
+
+from . import utils
 from . import version
 from . import v1version
 from . import v2version
@@ -31,9 +33,24 @@ PatternsByFile   = typ.Dict[str, typ.List[Pattern]]
 FilePatternsItem = typ.Tuple[str, typ.List[Pattern]]
 
 
-SUPPORTED_CONFIGS = ["setup.cfg", "pyproject.toml", "pycalver.toml", "bumpver.toml"]
+SUPPORTED_CONFIGS = [
+    "setup.cfg",
+    "pyproject.toml",
+    "pycalver.toml",
+    "bumpver.toml",
+    ".bumpver.toml",
+]
 
+
+class TagScope(str, enum.Enum):
+    DEFAULT = "default"
+    GLOBAL  = "global"
+    BRANCH  = "branch"
+
+
+DEFAULT_TAG_SCOPE      = TagScope.DEFAULT
 DEFAULT_COMMIT_MESSAGE = "bump version to {new_version}"
+DEFAULT_TAG_MESSAGE    = "{new_version}"
 
 
 class ProjectContext(typ.NamedTuple):
@@ -50,6 +67,7 @@ def _pick_config_filepath(path: pl.Path) -> pl.Path:
     config_candidates: typ.List[pl.Path] = [
         path / "pycalver.toml",
         path / "bumpver.toml",
+        path / ".bumpver.toml",
         path / "pyproject.toml",
         path / "setup.cfg",
     ]
@@ -61,9 +79,7 @@ def _pick_config_filepath(path: pl.Path) -> pl.Path:
             with config_filepath.open(mode="rb") as fobj:
                 data = fobj.read()
 
-            has_bumpver_section = (
-                b"bumpver]" in data or b"pycalver]" in data
-            ) and b"current_version" in data
+            has_bumpver_section = (b"bumpver]" in data or b"pycalver]" in data) and b"current_version" in data
             if has_bumpver_section:
                 return config_filepath
 
@@ -97,8 +113,8 @@ def init_project_ctx(project_path: typ.Union[str, pl.Path, None] = ".") -> Proje
     elif project_path is None:
         path = pl.Path(".")
     else:
-        # assume it's a str/unicode
-        path = pl.Path(project_path)
+        # assume it's coercable to str/unicode
+        path = pl.Path(str(project_path))
 
     config_filepath, config_rel_path, config_format = _parse_config_and_format(path)
 
@@ -125,6 +141,11 @@ class Config(typ.NamedTuple):
     version_pattern: str
     pep440_version : str
     commit_message : str
+    tag_message    : str
+    tag_scope      : TagScope
+
+    pre_commit_hook : str
+    post_commit_hook: str
 
     commit        : bool
     tag           : bool
@@ -144,6 +165,10 @@ def _debug_str(cfg: Config) -> str:
         f"\n    version_pattern='{cfg.version_pattern}',",
         f"\n    pep440_version='{cfg.pep440_version}',",
         f"\n    commit_message='{cfg.commit_message}',",
+        f"\n    tag_message='{cfg.tag_message}',",
+        f"\n    tag_scope='{cfg.tag_scope.value}',",
+        f"\n    pre_commit_hook={cfg.pre_commit_hook},",
+        f"\n    post_commit_hook={cfg.post_commit_hook},",
         f"\n    commit={cfg.commit},",
         f"\n    tag={cfg.tag},",
         f"\n    push={cfg.push},",
@@ -358,12 +383,21 @@ def _validate_version_with_pattern(
             raise ValueError(errmsg)
 
 
+def _parse_cfg_strings(raw_cfg: RawConfig, key: str, default: str) -> str:
+    if key in raw_cfg:
+        raw_cfg[key] = raw_cfg[key].strip("'\" ")
+    return raw_cfg.get(key, default)
+
+
 def _parse_config(raw_cfg: RawConfig) -> Config:
     """Parse configuration which was loaded from an .ini/.cfg or .toml file."""
 
     commit_message: str = raw_cfg.get('commit_message', DEFAULT_COMMIT_MESSAGE)
     commit_message = raw_cfg['commit_message'] = commit_message.strip("'\" ")
 
+    tag_message: str = raw_cfg.get('tag_message', DEFAULT_TAG_MESSAGE)
+    tag_message = raw_cfg['tag_message'] = tag_message.strip("'\" ")
+
     current_version: str = raw_cfg['current_version']
     current_version = raw_cfg['current_version'] = current_version.strip("'\" ")
 
@@ -378,6 +412,11 @@ def _parse_config(raw_cfg: RawConfig) -> Config:
 
     file_patterns = _compile_file_patterns(raw_cfg, is_new_pattern)
 
+    tag_scope: TagScope = TagScope(_parse_cfg_strings(raw_cfg, 'tag_scope', DEFAULT_TAG_SCOPE))
+
+    pre_commit_hook : str = _parse_cfg_strings(raw_cfg, 'pre_commit_hook' , "")
+    post_commit_hook: str = _parse_cfg_strings(raw_cfg, 'post_commit_hook', "")
+
     commit = raw_cfg['commit']
     tag    = raw_cfg['tag']
     push   = raw_cfg['push']
@@ -393,11 +432,26 @@ def _parse_config(raw_cfg: RawConfig) -> Config:
     if push and not commit:
         raise ValueError("commit=True required if push=True")
 
+    if tag_scope not in list(TagScope):
+        raise ValueError("Invalid value for tag_scope")
+
+    if pre_commit_hook and not pl.Path(pre_commit_hook).exists():
+        errmsg = f"Invalid value for pre_commit_hook: Path '{pre_commit_hook}' does not exist"
+        raise ValueError(errmsg)
+
+    if post_commit_hook and not pl.Path(post_commit_hook).exists():
+        errmsg = f"Invalid value for post_commit_hook: Path '{post_commit_hook}' does not exist"
+        raise ValueError(errmsg)
+
     cfg = Config(
         current_version=current_version,
         version_pattern=version_pattern,
         pep440_version=pep440_version,
         commit_message=commit_message,
+        tag_message=tag_message,
+        tag_scope=tag_scope,
+        pre_commit_hook=pre_commit_hook,
+        post_commit_hook=post_commit_hook,
         commit=commit,
         tag=tag,
         push=push,
@@ -500,6 +554,10 @@ DEFAULT_CONFIGPARSER_BASE_TMPL = """
 current_version = "{initial_version}"
 version_pattern = "YYYY.BUILD[-TAG]"
 commit_message = "bump version {{old_version}} -> {{new_version}}"
+tag_message = "{{new_version}}"
+tag_scope = "{default_tag_scope}"
+pre_commit_hook = ""
+post_commit_hook = ""
 commit = True
 tag = True
 push = True
@@ -540,6 +598,10 @@ DEFAULT_PYPROJECT_TOML_BASE_TMPL = """
 current_version = "{initial_version}"
 version_pattern = "YYYY.BUILD[-TAG]"
 commit_message = "bump version {{old_version}} -> {{new_version}}"
+tag_message = "{{new_version}}"
+tag_scope = "{default_tag_scope}"
+pre_commit_hook = ""
+post_commit_hook = ""
 commit = true
 tag = true
 push = true
@@ -553,6 +615,10 @@ DEFAULT_BUMPVER_TOML_BASE_TMPL = """
 current_version = "{initial_version}"
 version_pattern = "YYYY.BUILD[-TAG]"
 commit_message = "bump version {{old_version}} -> {{new_version}}"
+tag_message = "{{new_version}}"
+tag_scope = "{default_tag_scope}"
+pre_commit_hook = ""
+post_commit_hook = ""
 commit = true
 tag = true
 push = true
@@ -574,6 +640,11 @@ DEFAULT_TOML_BUMPVER_STR = """
 ]
 """.lstrip()
 
+DEFAULT_TOML_DOT_BUMPVER_STR = """
+".bumpver.toml" = [
+    'current_version = "{version}"',
+]
+""".lstrip()
 
 DEFAULT_TOML_PYPROJECT_STR = """
 "pyproject.toml" = [
@@ -607,11 +678,11 @@ DEFAULT_TOML_README_MD_STR = """
 
 
 def _initial_version() -> str:
-    return dt.datetime.utcnow().strftime("%Y.1001-alpha")
+    return utils.now().strftime("%Y.1001-alpha")
 
 
 def _initial_version_pep440() -> str:
-    return dt.datetime.utcnow().strftime("%Y.1001a0")
+    return utils.now().strftime("%Y.1001a0")
 
 
 def default_config(ctx: ProjectContext) -> str:
@@ -636,6 +707,7 @@ def default_config(ctx: ProjectContext) -> str:
             "pyproject.toml": DEFAULT_TOML_PYPROJECT_STR,
             "pycalver.toml" : DEFAULT_TOML_PYCALVER_STR,
             "bumpver.toml"  : DEFAULT_TOML_BUMPVER_STR,
+            ".bumpver.toml" : DEFAULT_TOML_DOT_BUMPVER_STR,
             "setup.py"      : DEFAULT_TOML_SETUP_PY_STR,
             "README.rst"    : DEFAULT_TOML_README_RST_STR,
             "README.md"     : DEFAULT_TOML_README_MD_STR,
@@ -643,7 +715,10 @@ def default_config(ctx: ProjectContext) -> str:
     else:
         raise ValueError(f"Invalid config_format='{fmt}', must be either 'toml' or 'cfg'.")
 
-    cfg_str = base_tmpl.format(initial_version=_initial_version())
+    cfg_str = base_tmpl.format(
+        initial_version=_initial_version(),
+        default_tag_scope=DEFAULT_TAG_SCOPE.value,
+    )
 
     for filename, default_str in default_pattern_strs_by_filename.items():
         if (ctx.path / filename).exists():
