@@ -1,9 +1,10 @@
 # This file is part of the bumpver project
 # https://github.com/mbarkhau/bumpver
 #
-# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
+# Copyright (c) 2018-2024 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
 # SPDX-License-Identifier: MIT
 import typing as typ
+import datetime as dt
 import functools
 
 # NOTE (mb 2020-09-24): The main use of the memo function is
@@ -22,3 +23,10 @@ def memo(func: typ.Callable) -> typ.Callable:
         return cache[key]
 
     return wrapper
+
+
+def now() -> dt.datetime:
+    if hasattr(dt, 'UTC'):
+        return dt.datetime.now(dt.UTC)
+    else:
+        return dt.datetime.utcnow()
