@@ -1,7 +1,7 @@
 # This file is part of the bumpver project
 # https://github.com/mbarkhau/bumpver
 #
-# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
+# Copyright (c) 2018-2024 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
 # SPDX-License-Identifier: MIT
 """Functions related to version string manipulation."""
 
@@ -35,7 +35,7 @@ def _is_cal_gt(left: CalInfo, right: CalInfo) -> bool:
     return lvals > rvals
 
 
-def cal_info(date: dt.date = None) -> version.V2CalendarInfo:
+def cal_info(date: typ.Optional[dt.date] = None) -> version.V2CalendarInfo:
     """Generate calendar components for current date.
 
     >>> import datetime as dt
@@ -224,6 +224,7 @@ def parse_field_values_to_vinfo(field_values: FieldValues) -> version.V2VersionI
     tag     = fvals.get('tag'    ) or ""
     pytag   = fvals.get('pytag'  ) or ""
     githash = fvals.get('githash') or ""
+    hexhash = fvals.get('hexhash') or ""
 
     if tag and not pytag:
         pytag = version.PEP440_TAG_BY_TAG[tag]
@@ -259,15 +260,14 @@ def parse_field_values_to_vinfo(field_values: FieldValues) -> version.V2VersionI
         tag=tag,
         pytag=pytag,
         githash=githash,
+        hexhash=hexhash,
         num=num,
         inc0=inc0,
         inc1=inc1,
     )
 
 
-def parse_version_info(
-    version_str: str, raw_pattern: str = "vYYYY0M.BUILD[-TAG]"
-) -> version.V2VersionInfo:
+def parse_version_info(version_str: str, raw_pattern: str = "vYYYY0M.BUILD[-TAG]") -> version.V2VersionInfo:
     """Parse normalized V2VersionInfo.
 
     >>> vinfo = parse_version_info("v201712.0033-beta", raw_pattern="vYYYY0M.BUILD[-TAG]")
@@ -661,14 +661,15 @@ def _reset_rollover_fields(
 
 
 def _incr_numeric(
-    raw_pattern: str,
-    old_vinfo  : version.V2VersionInfo,
-    cur_vinfo  : version.V2VersionInfo,
-    major      : bool,
-    minor      : bool,
-    patch      : bool,
-    tag        : typ.Optional[str],
-    tag_num    : bool,
+    raw_pattern   : str,
+    old_vinfo     : version.V2VersionInfo,
+    cur_vinfo     : version.V2VersionInfo,
+    major         : bool,
+    minor         : bool,
+    patch         : bool,
+    tag           : typ.Optional[str],
+    tag_num       : bool,
+    pin_increments: bool,
 ) -> version.V2VersionInfo:
     """Increment (and reset to zero) non CalVer parts.
 
@@ -685,6 +686,7 @@ def _incr_numeric(
     ...     patch=True,
     ...     tag='beta',
     ...     tag_num=False,
+    ...     pin_increments=False,
     ... )
     >>> (new_vinfo.major, new_vinfo.minor, new_vinfo.patch, new_vinfo.tag, new_vinfo.pytag, new_vinfo.num)
     (1, 2, 4, 'beta', 'b', 0)
@@ -704,8 +706,9 @@ def _incr_numeric(
         pytag     = version.PEP440_TAG_BY_TAG[tag]
         cur_vinfo = cur_vinfo._replace(pytag=pytag)
 
-    cur_vinfo = cur_vinfo._replace(inc0=cur_vinfo.inc0 + 1)
-    cur_vinfo = cur_vinfo._replace(inc1=cur_vinfo.inc1 + 1)
+    if not pin_increments:
+        cur_vinfo = cur_vinfo._replace(inc0=cur_vinfo.inc0 + 1)
+        cur_vinfo = cur_vinfo._replace(inc1=cur_vinfo.inc1 + 1)
 
     # prevent truncation of leading zeros
     if int(cur_vinfo.bid) < 1000:
@@ -738,13 +741,14 @@ def incr(
     old_version: str,
     raw_pattern: str = "vYYYY0M.BUILD[-TAG]",
     *,
-    major     : bool = False,
-    minor     : bool = False,
-    patch     : bool = False,
-    tag       : typ.Optional[str] = None,
-    tag_num   : bool = False,
-    pin_date  : bool = False,
-    maybe_date: typ.Optional[dt.date] = None,
+    major         : bool = False,
+    minor         : bool = False,
+    patch         : bool = False,
+    tag           : typ.Optional[str] = None,
+    tag_num       : bool = False,
+    pin_increments: bool = False,
+    pin_date      : bool = False,
+    maybe_date    : typ.Optional[dt.date] = None,
 ) -> typ.Optional[str]:
     """Increment version string.
 
@@ -784,11 +788,14 @@ def incr(
         patch=patch,
         tag=tag,
         tag_num=tag_num,
+        pin_increments=pin_increments,
     )
 
     new_version = format_version(cur_vinfo, raw_pattern)
 
-    if new_version == old_version:
+    if new_version == "":
+        return None
+    elif new_version == old_version:
         logger.error("Invalid arguments or pattern, version did not change.")
         return None
     else:
