@@ -1,7 +1,7 @@
 # This file is part of the bumpver project
 # https://github.com/mbarkhau/bumpver
 #
-# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
+# Copyright (c) 2018-2024 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
 # SPDX-License-Identifier: MIT
 #
 # bumpver/vcs.py (this file) is based on code from the
@@ -24,6 +24,7 @@ import logging
 import tempfile
 import subprocess as sp
 
+from . import hooks
 from . import config
 
 logger = logging.getLogger("bumpver.vcs")
@@ -44,27 +45,33 @@ BRANCH_RE = re.compile(BRANCH_PATTERN, flags=re.VERBOSE)
 
 VCS_SUBCOMMANDS_BY_NAME = {
     'git': {
-        'is_usable'   : "git rev-parse --git-dir",
-        'fetch'       : "git fetch",
-        'ls_tags'     : "git tag --list",
-        'status'      : "git status --porcelain",
-        'add_path'    : "git add --update '{path}'",
-        'commit'      : "git commit --message '{message}'",
-        'tag'         : "git tag --annotate {tag} --message {tag}",
-        'push_tag'    : "git push {remote} --follow-tags {tag} HEAD",
-        'show_remotes': "git config --get remote.origin.url",
-        'ls_branches' : "git branch -vv",
+        'is_usable'     : "git rev-parse --git-dir",
+        'fetch'         : "git fetch",
+        'ls_tags'       : "git tag --list",
+        'ls_tags_branch': "git tag --list --merged",
+        'status'        : "git status --porcelain",
+        'add_path'      : "git add --update '{path}'",
+        'commit'        : "git commit --message '{message}'",
+        'tag'           : "git tag --annotate {tag} --message '{message}'",
+        'tag_light'     : "git tag {tag}",
+        'push_tag'      : "git push {remote} --follow-tags {tag} HEAD",
+        'push'          : "git push {remote} HEAD",
+        'show_remotes'  : "git config --get remote.origin.url",
+        'ls_branches'   : "git branch -vv",
     },
     'hg': {
-        'is_usable'   : "hg root",
-        'fetch'       : "hg pull",
-        'ls_tags'     : "hg tags",
-        'status'      : "hg status -umard",
-        'add_path'    : "hg add {path}",
-        'commit'      : "hg commit --logfile {path}",
-        'tag'         : "hg tag {tag} --message {tag}",
-        'push_tag'    : "hg push {tag}",
-        'show_remotes': "hg paths",
+        'is_usable'     : "hg root",
+        'fetch'         : "hg pull",
+        'ls_tags'       : "hg tags",
+        'ls_tags_branch': "hg log --branch . --rev='tag()' --template='{{tags}}\\n'",
+        'status'        : "hg status -umard",
+        'add_path'      : "hg add '{path}'",
+        'commit'        : "hg commit --logfile '{path}'",
+        'tag'           : "hg tag {tag} --message '{message}'",
+        'tag_light'     : "hg tag {tag}",
+        'push_tag'      : "hg push {tag}",
+        'push'          : "hg push",
+        'show_remotes'  : "hg paths",
     },
 }
 
@@ -75,14 +82,14 @@ Env = typ.Dict[str, str]
 class VCSAPI:
     """Absraction for git and mercurial."""
 
-    def __init__(self, name: str, subcommands: typ.Dict[str, str] = None):
+    def __init__(self, name: str, subcommands: typ.Optional[typ.Dict[str, str]] = None):
         self.name = name
         if subcommands is None:
             self.subcommands = VCS_SUBCOMMANDS_BY_NAME[name]
         else:
             self.subcommands = subcommands
 
-    def __call__(self, cmd_name: str, env: Env = None, **kwargs: str) -> str:
+    def __call__(self, cmd_name: str, env: typ.Optional[Env] = None, **kwargs: str) -> str:
         """Invoke subcommand and return output."""
         cmd_tmpl = self.subcommands[cmd_name]
         cmd_str  = cmd_tmpl.format(**kwargs)
@@ -154,6 +161,12 @@ class VCSAPI:
         logger.debug(f"ls_tags output {ls_tag_lines}")
         return [line.strip().split(" ", 1)[0] for line in ls_tag_lines]
 
+    def ls_tags_branch(self) -> typ.List[str]:
+        """List vcs tags on all branches."""
+        ls_tag_lines = self('ls_tags_branch').splitlines()
+        logger.debug(f"ls_tags_branch output {ls_tag_lines}")
+        return [line.strip().split(" ", 1)[0] for line in ls_tag_lines]
+
     def add(self, path: str) -> None:
         """Add updates to be included in next commit."""
         try:
@@ -191,16 +204,27 @@ class VCSAPI:
             finally:
                 os.unlink(tmp_file.name)
 
-    def tag(self, tag_name: str) -> None:
-        """Create an annotated tag."""
-        self('tag', tag=tag_name)
+    def tag(self, tag_name: str, tag_message: str) -> None:
+        """Create a tag."""
+        if tag_message:
+            # Annotated
+            self('tag', tag=tag_name, message=tag_message)
+        else:
+            # Lightweight
+            self('tag_light', tag=tag_name)
 
-    def push(self, tag_name: str) -> None:
+    def push_tag(self, tag_name: str) -> None:
         """Push changes to origin."""
         remote = self.get_remote()
         if remote:
             self('push_tag', tag=tag_name, remote=remote)
 
+    def push(self) -> None:
+        """Push changes to origin."""
+        remote = self.get_remote()
+        if remote:
+            self('push', remote=remote)
+
     def __repr__(self) -> str:
         """Generate string representation."""
         return f"VCSAPI(name='{self.name}')"
@@ -247,28 +271,47 @@ def commit(
     filepaths     : typ.Set[str],
     new_version   : str,
     commit_message: str,
+    tag_message   : str,
 ) -> None:
     if cfg.commit:
+        if cfg.pre_commit_hook:
+            logger.info(f"Run pre-commit hook: {cfg.pre_commit_hook}")
+            hooks.run(cfg.pre_commit_hook, cfg.current_version, new_version)
+
         for filepath in filepaths:
             vcs_api.add(filepath)
 
         vcs_api.commit(commit_message)
 
+        if cfg.post_commit_hook:
+            logger.info(f"Run post-commit hook: {cfg.post_commit_hook}")
+            hooks.run(cfg.post_commit_hook, cfg.current_version, new_version)
+
     if cfg.commit and cfg.tag:
-        vcs_api.tag(new_version)
+        vcs_api.tag(tag_name=new_version, tag_message=tag_message)
 
     if cfg.commit and cfg.push:
-        vcs_api.push(new_version)
+        if cfg.tag:
+            vcs_api.push_tag(tag_name=new_version)
+        else:
+            vcs_api.push()
 
 
-def get_tags(fetch: bool) -> typ.List[str]:
+def get_tags(fetch: bool, scope: config.TagScope) -> typ.List[str]:
     try:
         vcs_api = get_vcs_api()
         logger.debug(f"vcs found: {vcs_api.name}")
+
         if fetch:
             logger.info("fetching tags from remote (to turn off use: -n / --no-fetch)")
             vcs_api.fetch()
-        return vcs_api.ls_tags()
+
+        branch_scope = scope == config.TagScope.BRANCH
+
+        if branch_scope:
+            return vcs_api.ls_tags_branch()
+        else:
+            return vcs_api.ls_tags()
     except OSError:
         logger.debug("No vcs found")
         return []
